global function SpawnDrivableDropShip
global function DEV_SpawnDrivableDropShip

global struct DropShiptruct
{
    ShipStruct& dropship
    entity mover
    entity camera
    float acceleration = 0.0
    float time_enter = 0.0
    float time_sound = 0.0
    float time_fired = 0.0
}

const float drop_ship_base_speed = 1000.0
const int base_rotation = 20
const flight_limit = 10000

/*
███████╗███████╗████████╗██╗   ██╗██████╗ 
██╔════╝██╔════╝╚══██╔══╝██║   ██║██╔══██╗
███████╗█████╗     ██║   ██║   ██║██████╔╝
╚════██║██╔══╝     ██║   ██║   ██║██╔═══╝ 
███████║███████╗   ██║   ╚██████╔╝██║     
╚══════╝╚══════╝   ╚═╝    ╚═════╝ ╚═╝     
                                          
*/

DropShiptruct function SpawnDrivableDropShip( vector origin, vector angles = CONVOYDIR, int team = TEAM_IMC )
{
    InitKeyTracking()

    DropShiptruct dropship

    ShipStruct ship = SpawnDropShipLight( WorldToLocalOrigin( origin ), angles, team, true )
    ship.behavior = eBehavior.CUSTOM
    // ship.behaviorTable[ eBehavior.DEATH_ANIM ] = Custom_DeathAnim

    ship.mover.SetPusher( true )
    ship.model.SetPusher( true )
    thread PlayAnim( ship.model, "dropship_open_doorL", dropship.dropship.mover )

    // CreatePropDynamic(  )
    ship.model.SetMaxHealth( 1500 * 2 )
	ship.model.SetHealth( ship.model.GetMaxHealth() )
    ship.model.SetUsableByGroup( "pilot" )
    ship.model.SetUsePrompts( "Hold %use% to embark", "Press %use% to embark" )
    

    ship.model.SetTitle( "dropship" )
    ship.model.SetScriptName( "drivable_dropship" )
    ShowName( ship.model )
    dropship.dropship = ship

    SetupRocketMarvin( dropship )

    thread DropShipWaitForDriver( dropship )
    
    return dropship
}

void function DropShipWaitForDriver( DropShiptruct dropship )
{
    entity model = dropship.dropship.model
    model.SetUsable()
    model.SetOwner( null )
    StopSoundOnEntity( model, "amb_emit_s2s_rushing_wind_strong_v2_02b" )

    entity player = expect entity( model.WaitSignal( "OnPlayerUse" ).player )

    model.UnsetUsable()

    if ( IsValid( model.GetOwner() ) || model.GetOwner() == player )
        return
    
    if ( !IsValid( player ) || player.IsTitan() || IsValid( player.GetParent() ) )
    {
        thread DropShipWaitForDriver( dropship )
        return 
    }
        
    
    dropship.time_enter = Time()
    SpawnDropshipCamera( player, dropship )
    player.SetOrigin( dropship.mover.GetOrigin() + <0,0,50> )
    player.SetParent( dropship.mover )
    player.MakeInvisible()
    player.SetInvulnerable()
    dropship.dropship.mover.SetOwner( player )
    ScreenFade( player, 0, 0, 0, 255, 0.3, 0.3, (FFADE_IN | FFADE_PURGE) )

    thread DropShipDrive( dropship )
}

void function DropShipDrive( DropShiptruct dropship )
{
    ShipStruct ship = dropship.dropship
    entity mover = ship.mover
    entity model = ship.model
    entity player = model.GetOwner()

    player.EndSignal( "OnDeath" )
    player.EndSignal( "OnDestroy" )
    model.EndSignal( "OverDamaged" )
    EndSignal( ship, "engineFailure_Complete" )
    model.EndSignal( "OnDeath" )
    model.EndSignal( "OnDestroy" )

    OnThreadEnd(
		function () : ( dropship, ship, model )
		{
			if ( !IsValid( model ) || model.GetHealth() < 10 || ship.behavior == eBehavior.DEATH_ANIM || ship.behavior == eBehavior.ENGINE_FAILURE )
                DropShipDied( dropship, model )
            else if ( !IsValid( model.GetOwner() ) )
                HandlePlayerDeathInDropship( dropship, model )
            else if ( IsValid( model.GetOwner() ) )
                HandleExitRequest( dropship, model.GetOwner() )
		}
	)

    vector angles
    vector attack_angles
    vector thing
    array< bool > keys

    for(;;)
    {
        // if ( ship.behavior == eBehavior.DEATH_ANIM )
        //     return
        
        thing = mover.GetOrigin()
        if ( thing.x > flight_limit || thing.x < -flight_limit || thing.y > flight_limit || thing.y < -flight_limit || thing.z > flight_limit - 3000 || thing.z < -flight_limit + 3000 )
            Signal( ship, "engineFailure_Complete" )


        keys = GetPlayerKeysList( player )

        if ( keys[KJ] )
        {
            IncreaseAcceleration( dropship )
        }
        else if ( keys[KD] ) 
        {
            DecreaseAcceleration( dropship )
        }

        if ( HasEnoughSpeed( dropship ) )
        {
            MoveDropShip( dropship, mover.GetForwardVector() )
        }

        
        if ( DidColide( dropship ) )
        {
            if ( GetAcceleration( dropship ) > 0.50 )
                model.TakeDamage( model.GetHealth() - 1, null, null, { damageSourceId=damagedef_suicide } )
            else
            {
                dropship.acceleration = -dropship.acceleration
                MoveDropShip( dropship, mover.GetForwardVector() * -300 )
                return
            }
        }
            
        
        angles = <0,0,0>
        attack_angles = mover.GetAngles()

        if ( keys[KL] )
        {
            angles.y = base_rotation
        }
        else if ( keys[KR] )
        {
            angles.y = -base_rotation
        }

        if ( keys[KB] && attack_angles.x > -50 )
        {
            // thing = mover.GetUpVector() * -base_rotation
            // angles += thing
            angles.x = -base_rotation
        }
        else if ( keys[KF] && attack_angles.x < 50 )
        {
            // thing = mover.GetUpVector() * base_rotation
            // angles += thing
            angles.x = base_rotation
        }

        RotateDropShip( dropship, angles )

        if ( keys[KA] )
            TryFireNuke( dropship )

        UpdatedCameraPosition( dropship )

        if ( dropship.time_sound < Time() ) // add hover and flight sounds
        {
            EmitSoundOnEntity( model, "amb_emit_s2s_rushing_wind_strong_v2_02b" )
            dropship.time_sound = Time() + 10
        }

        if ( dropship.time_enter < Time() - 2 && keys[KU] )
            return
        
        WaitFrame()
    }
}

void function DropShipDied( DropShiptruct dropship, entity model )
{
        vector origin = dropship.dropship.mover.GetOrigin()
        entity player = model.GetOwner()
        if ( IsValid( player ) )
        {
            DestroyDropShipCamera( dropship )
            player.MakeVisible()
            player.ClearParent()
            player.SetOrigin( origin + <0,0,150> )
            player.ClearInvulnerable()
            thread PlayerFlyOut( player )
        }

        if ( IsValid( dropship.dropship.pilot ) )
            dropship.dropship.pilot.Destroy() // or .Die()
}

void function PlayerFlyOut( entity player )
{
    player.EndSignal( "OnDeath" )
    player.EndSignal( "OnDestroy" )

    while( !player.IsOnGround() && IsValid( player ) && IsAlive( player ) )
    {
        vector thing = player.GetOrigin()
        if ( thing.x > flight_limit || thing.x < -flight_limit || thing.y > flight_limit || thing.y < -flight_limit || thing.z > flight_limit - 3000 || thing.z < -flight_limit + 3000 )
            player.Die()
        
        WaitFrame()
    }
}

void function HandlePlayerDeathInDropship( DropShiptruct dropship, entity model )
{
    DestroyDropShipCamera( dropship )
    // model.GetOwner().ClearParent()

    vector angles = dropship.dropship.mover.GetAngles()
    angles.x = 0
    // dropship.dropship.mover.NonPhysicsRotateTo( angles, 2, 0.1, 0.1 )
    dropship.dropship.mover.SetAngles( angles )
    dropship.dropship.model.SetAngles( angles )

    thread DropShipWaitForDriver( dropship )

    thread SelfDriveThink( dropship )
}

void function HandleExitRequest( DropShiptruct dropship, entity player )
{
    DestroyDropShipCamera( dropship )
    player.MakeVisible()
    player.SetOrigin( dropship.dropship.model.GetOrigin() + <0,0,150> )
    player.ClearParent()
    player.ClearInvulnerable()
    thread PlayerFlyOut( player )

    // dropship.camera.Destroy()

    vector angles = dropship.dropship.mover.GetAngles()
    angles.x = 0
    dropship.dropship.mover.SetAngles( angles )
    dropship.dropship.model.SetAngles( angles )

    thread DropShipWaitForDriver( dropship )

    thread SelfDriveThink( dropship )
}

void function SelfDriveThink( DropShiptruct dropship )
{
    ShipStruct ship = dropship.dropship
    entity mover = ship.mover
    entity model = ship.model

    while( !IsValid( model.GetOwner() ) )
    {
        if ( !HasEnoughSpeed( dropship ) )
            return
        
        if ( DidColide( dropship ) )
            model.TakeDamage( model.GetHealth() - 1, null, null, { damageSourceId=damagedef_suicide } )

        if ( GetAcceleration( dropship ) < 0.0 )
        {
            IncreaseAcceleration( dropship )
        }
        else if ( GetAcceleration( dropship ) > 0.0 )
        {
            DecreaseAcceleration( dropship )
        }

        MoveDropShip( dropship, mover.GetForwardVector() )

        WaitFrame()
    }
}

/*
██╗   ██╗████████╗██╗██╗     ██╗████████╗██╗███████╗███████╗
██║   ██║╚══██╔══╝██║██║     ██║╚══██╔══╝██║██╔════╝██╔════╝
██║   ██║   ██║   ██║██║     ██║   ██║   ██║█████╗  ███████╗
██║   ██║   ██║   ██║██║     ██║   ██║   ██║██╔══╝  ╚════██║
╚██████╔╝   ██║   ██║███████╗██║   ██║   ██║███████╗███████║
 ╚═════╝    ╚═╝   ╚═╝╚══════╝╚═╝   ╚═╝   ╚═╝╚══════╝╚══════╝                                                                                              
*/

bool function DidColide( DropShiptruct dropship )
{
    TraceResults traceResult
    vector Ray
    foreach( float angle in [10.0,0.0,-10.0] )
    {
        Ray = AnglesToForward( dropship.dropship.mover.GetAngles() + <0,angle,0>, <0,0,0> ) * 80
        traceResult = TraceLine( dropship.dropship.mover.GetOrigin(), dropship.dropship.mover.GetOrigin() + Ray, [ dropship.dropship.mover, dropship.dropship.model ], TRACE_MASK_SHOT, TRACE_COLLISION_GROUP_NONE )

        // print( "b" + traceResult.hitEnt )
        
        if ( traceResult.hitEnt != null )
        {

            if ( traceResult.hitEnt.IsPlayer() || traceResult.hitEnt.IsNPC() )
                continue
            
            print( "DropShip Collided with " + traceResult.hitEnt.GetScriptName() + ", " + traceResult.hitEnt.GetClassName() )
            return true
        }
    }

    return false
}

void function MoveDropShip( DropShiptruct dropship, vector offset )
{
    dropship.dropship.mover.NonPhysicsMoveTo( dropship.dropship.mover.GetOrigin() + ( offset  * ( drop_ship_base_speed * dropship.acceleration ) ), 0.3, 0.05, 0.05 )
}

void function RotateDropShip( DropShiptruct dropship, vector angles )
{
    dropship.dropship.mover.NonPhysicsRotateTo( dropship.dropship.mover.GetAngles() + angles, 0.3, 0.05, 0.05 )
}

bool function HasEnoughSpeed( DropShiptruct dropship )
{
    if ( dropship.acceleration > 0.25 || dropship.acceleration < -0.25 )
        return true
    
    return false
}

float function GetAcceleration( DropShiptruct dropship )
{
    return dropship.acceleration
}

void function IncreaseAcceleration( DropShiptruct dropship )
{
    if ( dropship.acceleration < 1.0 )
    {
        dropship.acceleration += 0.02
    }
    else
    {
        dropship.acceleration = 1.0
    }
}
void function DecreaseAcceleration( DropShiptruct dropship )
{
    if ( dropship.acceleration > 0.0 )
    {
        dropship.acceleration -= 0.02
    }
    else
    {
        dropship.acceleration = 0.0
    }
}

void function SpawnDropshipCamera( entity player, DropShiptruct dropship )
{
    entity model = dropship.dropship.model

    player.SnapEyeAngles( model.GetAngles() )
    // entity viewControl = CreateEntity( "point_viewcontrol" )
	// viewControl.kv.spawnflags = 56 // infinite hold time, snap to goal angles, make player non-solid

	// viewControl.SetOrigin( model.GetOrigin() )
	// viewControl.SetAngles( model.GetAngles() )
	// DispatchSpawn( viewControl )
    
    // EntFireByHandle( viewControl, "Enable", "!activator", 0, player, null )

    entity mover = CreateExpensiveScriptMover()
    mover.SetOrigin( model.GetOrigin() )
	mover.SetAngles( model.GetAngles() )
    // viewControl.SetParent( mover )

    player.SnapEyeAngles( model.GetAngles() )

    dropship.mover = mover
    // dropship.camera = viewControl
    
    player.SetAngles( dropship.dropship.model.GetAngles() )
    HolsterAndDisableWeapons( player )
    player.PlayerCone_SetLerpTime( 0.5 )
    player.PlayerCone_FromAnim()
	player.PlayerCone_SetMinYaw( 0 )
	player.PlayerCone_SetMaxYaw( 0 )
	player.PlayerCone_SetMinPitch( 0 )
	player.PlayerCone_SetMaxPitch( 0 )
}

void function DestroyDropShipCamera( DropShiptruct dropship )
{
    entity player = dropship.dropship.model.GetOwner()
    if ( IsValid( player ) )
    {
        // EntFireByHandle( dropship.camera, "Disable", "!activator", 0, player, null )
        player.ClearParent()
        DeployAndEnableWeapons( player )
        ViewConeFree( player )
    }

    // dropship.camera.Destroy()
    dropship.mover.Destroy()
}

void function UpdatedCameraPosition( DropShiptruct dropship )
{
    entity mover = dropship.dropship.mover
    entity player = dropship.dropship.model.GetOwner()
    vector offset = mover.GetForwardVector() * -500
    offset += mover.GetUpVector() * 200

    dropship.mover.NonPhysicsMoveTo( offset + mover.GetOrigin(), 0.3, 0.05, 0.05 )
	dropship.mover.NonPhysicsRotateTo( mover.GetAngles(), 0.3, 0.05, 0.05 )

    // player.SnapEyeAngles( dropship.mover.GetAngles() )
}

void function DEV_SpawnDrivableDropShip()
{
    entity player = GetPlayerArray()[0]
    SpawnDrivableDropShip( player.GetOrigin(), CONVOYDIR, player.GetTeam() )
}